using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Design;
using System.Data;
using System.Text;
using System.Windows.Forms;
using System.Windows.Forms.Layout;

using MySQL.Utilities.Properties;
using MySQL.Utilities.SysUtils;

namespace MySQL.Utilities
{
  #region Delegates

  public delegate int TabCountHandler(object sender);
  public delegate bool TabInfoHandler(object sender, int index,
    out int iconId, out string caption, out string description);
  public delegate bool TabChangingHandler(object sender, int index);
  public delegate void TabChangedHandler(object sender, int index);
  public delegate void CustomButtonEvent(object sender, int index);

  #endregion

  public enum CollapsingPanelDisplayMode { Normal, HeaderAndTab, TabsOnly, HeaderWithTabs };
  public enum CollapsingPanelStyle { Convex, Flat };

  public class CollapsingPanel : Panel
  {
    #region Local classes

    // Holds data for a single custom button on the header.
    private class CustomButton
    {
      public Rectangle boundingBox;
      public Bitmap bitmap;

      public CustomButton(Bitmap bitmap, Rectangle bounds)
      {
        this.bitmap = bitmap;
        boundingBox = bounds;
      }

      public void Paint(Graphics graphics, bool active)
      {
        int x = boundingBox.Left;
        int y = boundingBox.Top;

        // Draw background first if the button is active.
        if (active)
        {
          Bitmap background = Resources.collapsing_panel_grid_bg;
          int bgX = x + (boundingBox.Width - background.Width) / 2;
          int bgY = y + (boundingBox.Height - background.Height) / 2;
          graphics.DrawImageUnscaled(background, bgX, bgY);

          // Move the button image down and right a bit to create the "pressed" effect.
          x++;
          y++;
        }

        // Center images in the available space.
        x += (boundingBox.Width - bitmap.Width) / 2;
        y += (boundingBox.Height - bitmap.Height) / 2;
        graphics.DrawImageUnscaled(bitmap, x, y);

        // Draw a left border line if not active.
        if (!active)
        {
          Bitmap boundary = Resources.collapsing_panel_header_tab_left_flat;
          graphics.DrawImageUnscaledAndClipped(boundary, new Rectangle(x, y, boundary.Width, boundingBox.Height));
        }
      }
    }

    #endregion

    // ----------------------------------------------------------------------------------------------------------------

    #region Member Variables

    /// <summary> 
    /// Required designer variable.
    /// </summary>
    private System.ComponentModel.IContainer components = null;

    // Drawing resources.
    private Brush fontDarkBrush = new SolidBrush(Color.FromArgb(30, 30, 30));

    private Font headerFont = new Font("Trebuchet MS", 11, FontStyle.Bold);
    private Font tabHeaderCaptionFont = new Font("Trebuchet MS", 11, FontStyle.Bold);
    private Font tabHeaderDescriptionFont = new Font("Trebuchet MS", 7);

    // Custom buttons. They are painted in the header, right aligned, left to the +/- buttons,
    // from left to right (index 0 is at the left side).
    private List<CustomButton> customButtons = new List<CustomButton>();
    private CustomButton activeCustomButton = null;

    // General properties.
    private string headerCaption = "Caption";

    private bool expanded = true;
    private int computedExpandedHeight = 0;
    private int manualHeight = 0;
    private int headerHeight = 23;
    private int headerHeightTabsOnly = 63;
    private int headerSpace = 5;
    private int buttonOffset = 0;

    private CollapsingPanelDisplayMode displayMode = CollapsingPanelDisplayMode.Normal;
    private CollapsingPanelStyle style = CollapsingPanelStyle.Flat;
    private bool disableExpansion = false;

    private int tabsHeight = 56;
    private int selectedTabIndex = 1;
    private ImageList tabHeaderImageList;
    private List<int> cachedTabWidths = new List<int>();
    private int tabCount = 0;
    private bool doPreviousControlResize = false;
    private Point mouseDownPoint;
    private int prevControlHeight;
    private bool displayTabActionButtons = false;
    private bool displayCustomButtons = false;
    private bool resizable = false;

    private ColumnLayout layoutEngine;

    #endregion

    #region Constructors

    public CollapsingPanel()
    {
      InitializeComponent();

      // Set double buffer
      SetStyle(
            ControlStyles.UserPaint |
            ControlStyles.AllPaintingInWmPaint |
            ControlStyles.OptimizedDoubleBuffer, true);
    }

    public CollapsingPanel(string Caption,
      CollapsingPanelDisplayMode InitialDisplayMode,
      CollapsingPanelStyle InitialStyle, bool InitialExpandState, int initialHeight)
      : this()
    {
      headerCaption = Caption;

      displayMode = InitialDisplayMode;
      style = InitialStyle;
      manualHeight = initialHeight;
      Expanded = InitialExpandState;
    }

    /// <summary> 
    /// Clean up any resources being used.
    /// </summary>
    /// <param name="disposing">true if managed resources should be disposed; otherwise, false.</param>
    protected override void Dispose(bool disposing)
    {
      if (disposing && (components != null))
      {
        components.Dispose();
      }
      base.Dispose(disposing);
    }

    #endregion

    #region Component Designer generated code

    /// <summary> 
    /// Required method for Designer support - do not modify 
    /// the contents of this method with the code editor.
    /// </summary>
    private void InitializeComponent()
    {
      components = new System.ComponentModel.Container();
    }

    #endregion

    #region Events

    protected override void OnControlAdded(ControlEventArgs e)
    {
      base.OnControlAdded(e);
      PerformLayout();
    }

    protected override void OnControlRemoved(ControlEventArgs e)
    {
      base.OnControlRemoved(e);
      PerformLayout();
    }

    protected override void OnPaint(System.Windows.Forms.PaintEventArgs e)
    {
      base.OnPaint(e);

      Graphics g = e.Graphics;

      // Draw the header background
      if (DisplayMode != CollapsingPanelDisplayMode.TabsOnly)
        if (style != CollapsingPanelStyle.Flat ||
          DisplayMode == CollapsingPanelDisplayMode.HeaderWithTabs)
        {
          Bitmap headerBg = style == CollapsingPanelStyle.Convex ?
            Resources.collapsing_panel_header_bg : Resources.collapsing_panel_header_bg_flat;

          if (headerHeight < headerBg.Height)
            headerHeight = headerBg.Height;

          g.DrawImageUnscaledAndClipped(headerBg,
            new Rectangle(0, 0, Width + 30, headerBg.Height));
        }
        else
        {
          g.FillRectangle(new SolidBrush(Color.FromArgb(235, 235, 235)),
            new Rectangle(0, 0, Width, headerHeight));
          g.DrawLine(new Pen(Color.FromArgb(180, 180, 180)),
            0, headerHeight - 1, Width, headerHeight - 1);

          // Draw tab action buttons
          if (displayTabActionButtons)
          {
            g.DrawImageUnscaledAndClipped(Resources.collapsing_panel_header_tab_add,
              new Rectangle(Width - Resources.collapsing_panel_header_tab_add.Width -
                  Resources.collapsing_panel_header_tab_del.Width,
                0,
                Resources.collapsing_panel_header_tab_add.Width,
                Resources.collapsing_panel_header_tab_add.Height));

            g.DrawImageUnscaledAndClipped(Resources.collapsing_panel_header_tab_del,
              new Rectangle(Width - Resources.collapsing_panel_header_tab_del.Width,
                0,
                Resources.collapsing_panel_header_tab_del.Width,
                Resources.collapsing_panel_header_tab_del.Height));
          }

          // Draw tab action buttons
          if (displayCustomButtons)
          {
            foreach (CustomButton button in customButtons)
              button.Paint(g, activeCustomButton == button);
          }
        }

      // If the header is expanded
      if (Expanded)
      {
        // Draw the Minus icon
        if (!disableExpansion && DisplayMode != CollapsingPanelDisplayMode.TabsOnly)
          g.DrawImageUnscaled(style == CollapsingPanelStyle.Convex ?
            Resources.collapsing_panel_minus : Resources.collapsing_panel_minus_flat, 10, 6);

        // Draw shadow
        if (DisplayMode != CollapsingPanelDisplayMode.TabsOnly)
        {
          // If the tabs are not displayed, draw the background of the shadow
          if (DisplayMode != CollapsingPanelDisplayMode.HeaderAndTab &&
            DisplayMode != CollapsingPanelDisplayMode.TabsOnly &&
            style != CollapsingPanelStyle.Flat)
          {
            g.FillRectangle(SystemBrushes.ButtonFace, 0, headerHeight,
              Width, headerHeight + Padding.Top + headerSpace - headerHeight);
          }
          else if (DisplayMode == CollapsingPanelDisplayMode.HeaderAndTab)
          {
            // If there are tabs below the header, draw their background
            g.FillRectangle(style == CollapsingPanelStyle.Convex ? SystemBrushes.ButtonFace :
              new SolidBrush(Color.FromArgb(236, 242, 248)),
              0, headerHeight, Width, headerHeight + Padding.Top + tabsHeight - headerHeight);
            g.DrawLine(SystemPens.ControlDark, 0, headerHeight + Padding.Top + tabsHeight - 1,
              Width, headerHeight + Padding.Top + tabsHeight - 1);
          }

          // Draw shadow image
          if (style != CollapsingPanelStyle.Flat)
          {
            Rectangle shadowRect = new Rectangle(0, headerHeight,
              Width + 30, Resources.collapsing_panel_header_shadow.Height);
            g.DrawImageUnscaledAndClipped(Resources.collapsing_panel_header_shadow,
              shadowRect);
            //for (int i = 0; i < Width; i += Resources.collapsing_panel_header_shadow.Width)
            //  g.DrawImage(Resources.collapsing_panel_header_shadow, i, headerHeight);
          }
        }
      }
      else
        // Draw the Plus icon
        if (!disableExpansion && DisplayMode != CollapsingPanelDisplayMode.TabsOnly)
          g.DrawImageUnscaled(style == CollapsingPanelStyle.Convex ?
            Resources.collapsing_panel_plus : Resources.collapsing_panel_plus_flat, 10, 6);


      // If tabs should be displayed
      if ((Expanded && (DisplayMode == CollapsingPanelDisplayMode.HeaderAndTab ||
          DisplayMode == CollapsingPanelDisplayMode.TabsOnly)) ||
          DisplayMode == CollapsingPanelDisplayMode.HeaderWithTabs)
      {
        // Draw the tab header background for TabsOnly
        if (DisplayMode == CollapsingPanelDisplayMode.TabsOnly)
        {
          g.FillRectangle(style == CollapsingPanelStyle.Convex ? SystemBrushes.ButtonFace :
            new SolidBrush(Color.FromArgb(236, 242, 248)),
            0, 0, Width, headerHeightTabsOnly);

          g.DrawLine(SystemPens.ControlDark, 0, headerHeightTabsOnly - 1,
            Width, headerHeightTabsOnly - 1);
          g.DrawImageUnscaledAndClipped(Resources.collapsing_panel_bg,
            new Rectangle(0, 0, Width + 30, Resources.collapsing_panel_bg.Height));
        }

        // Get the current number of tabs
        if (TabCountNeeded != null)
          tabCount = TabCountNeeded(this);
        else
          tabCount = 2;

        cachedTabWidths.Clear();

        // The X position of the current item
        int itemPos = (DisplayMode == CollapsingPanelDisplayMode.HeaderWithTabs) ?
          (disableExpansion ? 0 : 32) : 8;
        // The Y helper position
        int y = (DisplayMode == CollapsingPanelDisplayMode.TabsOnly) ?
          7 : headerHeight + Padding.Top;

        // Loop over the tab headers and draw them
        for (int i = 0; i < tabCount; i++)
        {
          int x = itemPos;
          int itemWidth = 0;
          string caption = "", description = "";
          int iconId = -1;
          bool drawIcon = false;
          float captionWidth = 0, descriptionWidth = 0;

          try
          {
            if (TabInfoNeeded != null)
            {
              if (!TabInfoNeeded(this, i + 1, out iconId, out caption, out description))
                continue;
            }
            else
            {
              caption = "Caption";
              description = "Description";
            }

            if (DisplayMode == CollapsingPanelDisplayMode.HeaderWithTabs)
            {
              // Get caption width
              captionWidth = g.MeasureString(caption, HeaderFont).Width;

              // Check if the is the selected tab
              if (i == selectedTabIndex - 1)
              {
                Bitmap tabLeft = style == CollapsingPanelStyle.Convex ?
                  Resources.collapsing_panel_header_tab_left :
                  Resources.collapsing_panel_header_tab_left_flat;
                Bitmap tabMiddle = style == CollapsingPanelStyle.Convex ?
                  Resources.collapsing_panel_header_tab_middle :
                  Resources.collapsing_panel_header_tab_middle_flat;
                Bitmap tabRight = style == CollapsingPanelStyle.Convex ?
                  Resources.collapsing_panel_header_tab_right :
                  Resources.collapsing_panel_header_tab_right_flat;

                if (DisableExpansion && i > 0)
                {
                  g.DrawImageUnscaled(tabLeft, x, 0);
                  x += tabLeft.Width;
                  itemWidth += tabLeft.Width;
                }

                g.DrawImageUnscaledAndClipped(tabMiddle, new Rectangle(x, 0,
                  Convert.ToInt16(captionWidth) + 16, tabMiddle.Height));

                x += 8;
                itemWidth += 8;

                // Draw the header caption
                if (style == CollapsingPanelStyle.Convex)
                  g.DrawString(caption, HeaderFont, Brushes.Black,
                    x, Expanded ? 3 : 1);
                else
                {
                  g.DrawString(caption, HeaderFont, fontDarkBrush,
                    x, 2);
                  g.DrawString(caption, HeaderFont, Brushes.White,
                    x, 1);
                }

                x += Convert.ToInt16(captionWidth) + 8;
                itemWidth += Convert.ToInt16(captionWidth) + 8;

                g.DrawImageUnscaled(tabRight, x, 0);

                itemWidth += tabRight.Width;
              }
              // if the tab is not selected
              else
              {
                Bitmap sep = style == CollapsingPanelStyle.Convex ?
                  Resources.collapsing_panel_header_tab_separator :
                  Resources.collapsing_panel_header_tab_separator_flat;

                // Draw first separator
                if (!DisableExpansion && i == 0)
                {
                  g.DrawImageUnscaled(sep, x, 0);

                  itemWidth = sep.Width;
                  x += itemWidth;
                }

                // space to last separator
                itemWidth += 8;
                x += 8;

                // Draw the header caption
                g.DrawString(caption, HeaderFont, style == CollapsingPanelStyle.Convex ?
                  Brushes.White : fontDarkBrush, x, 1);

                // Add caption width
                x += Convert.ToInt16(captionWidth) + 8;
                itemWidth += Convert.ToInt16(captionWidth) + 8;

                // Draw ending separator only if the next tab is not selected
                if (i != selectedTabIndex - 2)
                {
                  g.DrawImageUnscaled(sep, x, 0);

                  itemWidth += sep.Width;
                }
              }
            }
            else
            {
              // Check if an icon is assigned to this tab
              if (tabHeaderImageList != null &&
                iconId >= 0 && iconId < tabHeaderImageList.Images.Count)
              {
                drawIcon = true;

                // Add icon width and spacing
                itemWidth = tabHeaderImageList.ImageSize.Width + 16;
              }

              // Get caption and description width
              if (!caption.Equals(""))
                captionWidth = g.MeasureString(caption, tabHeaderCaptionFont).Width;

              if (!description.Equals(""))
                descriptionWidth = g.MeasureString(description, tabHeaderDescriptionFont).Width;

              // Add the longer text with to the itemWidth
              itemWidth += (int)Math.Ceiling(Math.Max(captionWidth, descriptionWidth));

              // If there was a text at all, add spacing
              if (captionWidth + descriptionWidth > 0)
                if (drawIcon)
                  itemWidth += 6;
                else
                  itemWidth += 16;

              // if there is nothing to draw, continue
              if (itemWidth == 0)
                continue;

              // Check if the is the selected tab
              if (i == selectedTabIndex - 1)
              {
                // Get rectangle for the middle image
                Rectangle middleRect = new Rectangle(
                  itemPos + Resources.collapsing_panel_tab_left.Width,
                  y, itemWidth -
                    Resources.collapsing_panel_tab_left.Width -
                    Resources.collapsing_panel_tab_right.Width + 1,
                    Resources.collapsing_panel_tab_middle.Height);

                // Draw the left border, middle image and right border
                g.DrawImageUnscaled(Resources.collapsing_panel_tab_left, itemPos,
                  y);
                g.DrawImageUnscaledAndClipped(Resources.collapsing_panel_tab_middle,
                  middleRect);
                g.DrawImage(Resources.collapsing_panel_tab_right,
                  itemPos + itemWidth - Resources.collapsing_panel_tab_right.Width,
                  y);
              }


              // indent for the icon or text
              x += 8;

              // Draw the icon
              if (drawIcon)
              {
                g.DrawImageUnscaled(tabHeaderImageList.Images[iconId],
                  x, ((DisplayMode == CollapsingPanelDisplayMode.TabsOnly) ? 0 : headerHeight + Padding.Top) +
                    (((DisplayMode == CollapsingPanelDisplayMode.TabsOnly) ? headerHeightTabsOnly : tabsHeight) -
                    tabHeaderImageList.Images[iconId].Height) / 2);

                x += tabHeaderImageList.ImageSize.Width + 6;
              }

              // Draw the caption
              if (!caption.Equals(""))
                g.DrawString(caption, tabHeaderCaptionFont, Brushes.Black,
                  x, y + 15);
              // Draw the description
              if (!description.Equals(""))
                g.DrawString(description, tabHeaderDescriptionFont, Brushes.Gray,
                  x, y + 15 + 16);
            }

            itemPos += itemWidth;
          }
          finally
          {
            cachedTabWidths.Add(itemWidth);
          }
        }

        // Draw the background of the shadow below the tab header
        SolidBrush b = new SolidBrush(BackColor);
        g.FillRectangle(b, 0, headerHeight + Padding.Top + tabsHeight, Width, headerSpace);
      }

      if (DisplayMode != CollapsingPanelDisplayMode.TabsOnly &&
        DisplayMode != CollapsingPanelDisplayMode.HeaderWithTabs)
      {
        float h = g.MeasureString(HeaderCaption, HeaderFont).Height;

        // Draw the header caption
        g.DrawString(HeaderCaption, HeaderFont,
          style == CollapsingPanelStyle.Convex ? Brushes.White : fontDarkBrush,
          (disableExpansion ? 6 : 30), Convert.ToInt16((headerHeight - h) / 2));
      }
    }

    private Control GetPreviousControl()
    {
      if (Parent != null)
      {
        int i = Parent.Controls.IndexOf(this);
        if (i < Parent.Controls.Count - 1)
        {
          Control c = Parent.Controls[i + 1];

          while (c is CollapsingPanel && !(c as CollapsingPanel).Expanded &&
            i < Parent.Controls.Count - 2)
            c = Parent.Controls[++i + 1];

          if (!(c is CollapsingPanel) || (c as CollapsingPanel).Expanded)
            return c;
        }
      }

      return null;
    }

    protected override void OnMouseDown(MouseEventArgs e)
    {
      if (resizable && (e.Button == MouseButtons.Left) && (e.X > 10 + 12) && (e.X < buttonOffset) && 
        (e.Y < headerHeight / 3))
      {
        Control c = GetPreviousControl();
        if (c != null)
        {
          doPreviousControlResize = true;
          mouseDownPoint = this.PointToScreen(e.Location);
          prevControlHeight = c.Height;
        }
      }

      base.OnMouseDown(e);
    }

    protected override void OnMouseMove(MouseEventArgs e)
    {
      if (resizable && (e.Y < headerHeight / 3))
      {
        if ((e.X >= 10 + 12 || DisableExpansion)
          && (TabAddButtonClicked == null || (e.X < buttonOffset))
          && GetPreviousControl() != null)
        {
          if (Cursor != Cursors.HSplit)
            Cursor = Cursors.HSplit;
        }
        else
          Cursor = Cursors.Default;
      }
      else
        Cursor = Cursors.Default;

      if (doPreviousControlResize && e.Button == MouseButtons.Left)
      {
        Control c = GetPreviousControl();
        if (c != null)
        {
          c.Height = prevControlHeight + this.PointToScreen(e.Location).Y - mouseDownPoint.Y;
          c.Update();
        }
      }

      base.OnMouseMove(e);
    }

    protected override void OnMouseUp(MouseEventArgs e)
    {
      if (doPreviousControlResize)
        doPreviousControlResize = false;

      if (TabHeaderMouseUp != null && e.Button == MouseButtons.Right)
      {
        if (GetTabAtPosition(e.X, e.Y) >= 0)
          TabHeaderMouseUp(this, e);
      }

      base.OnMouseUp(e);
    }

    protected override void OnMouseClick(MouseEventArgs e)
    {
      bool handled = false;

      if (e.Button == MouseButtons.Left)
      {
        // Change the expanded state only when header is not hidden
        if (!disableExpansion && DisplayMode != CollapsingPanelDisplayMode.TabsOnly
          && e.Y < headerHeight)
        {
          if (e.X <= 10 + 12)
          {
            Expanded = !Expanded;
            handled = true;
          }
          else
          {
            int offsetX = Width;

            // Test for - button hit.
            if (displayTabActionButtons)
            {
              offsetX -= Resources.collapsing_panel_header_tab_del.Width;
              if (e.X > offsetX && TabDelButtonClicked != null)
              {
                TabDelButtonClicked(this, new EventArgs());
                handled = true;
              }
              else
              {
                // + button hit.
                offsetX -= Resources.collapsing_panel_header_tab_add.Width;
                if (e.X > offsetX && TabAddButtonClicked != null)
                {
                  TabAddButtonClicked(this, new EventArgs());
                  handled = true;
                }
              }
            }

            if (!handled)
            {
              // Any custom button.
              if (CustomButtonClicked != null)
              {
                for (int index = customButtons.Count - 1; index >= 0; index--)
                {
                  CustomButton button = customButtons[index];
                  offsetX -= button.boundingBox.Width;
                  if (e.X > offsetX)
                  {
                    if (activeCustomButton != null)
                      Invalidate(activeCustomButton.boundingBox);

                    activeCustomButton = button;
                    CustomButtonClicked(this, index);
                    handled = true;
                    Invalidate(activeCustomButton.boundingBox);

                    break;
                  }
                }
              }
            }
          }
        }

        if (!handled)
        {
          // If header is hidden, the full Y range counts for tab selections
          if (DisplayMode == CollapsingPanelDisplayMode.TabsOnly ||
            DisplayMode == CollapsingPanelDisplayMode.HeaderWithTabs ||
            (e.Y > headerHeight + Padding.Top && e.Y < headerHeight + Padding.Top + tabsHeight))
          {
            int x = (DisplayMode == CollapsingPanelDisplayMode.HeaderWithTabs) ?
              (disableExpansion ? 0 : 32) : 8;

            for (int i = 0; i < tabCount; i++)
            {
              if (i >= cachedTabWidths.Count)
                break;

              if (cachedTabWidths[i] == 0)
                continue;

              if (e.X >= x && e.X < x + cachedTabWidths[i])
              {
                SelectedTabIndex = i + 1;
                handled = true;
                break;
              }

              x += cachedTabWidths[i];
            }
          }
        }
      }

      if (!handled)
        base.OnMouseClick(e);
    }

    protected override void OnMouseDoubleClick(MouseEventArgs e)
    {
      if (!disableExpansion && e.Button == MouseButtons.Left)
      {
        if (e.Y < headerHeight && e.X < buttonOffset)
          Expanded = !Expanded;
      }

      if (e.Button == MouseButtons.Left &&
        DisplayMode == CollapsingPanelDisplayMode.HeaderAndTab ||
        DisplayMode == CollapsingPanelDisplayMode.HeaderWithTabs ||
        DisplayMode == CollapsingPanelDisplayMode.TabsOnly)
      {
        int y1 = 0, y2 = 0;

        if (DisplayMode == CollapsingPanelDisplayMode.HeaderAndTab)
        {
          y1 = headerHeight;
          y2 = headerHeight + tabsHeight;
        }
        else if (DisplayMode == CollapsingPanelDisplayMode.HeaderWithTabs)
          y2 = headerHeight;
        else if (DisplayMode == CollapsingPanelDisplayMode.TabsOnly)
          y2 = tabsHeight;

        if (e.Y >= y1 && e.Y < y2 && TabDoubleClicked != null)
          TabDoubleClicked(this, e);
      }

      base.OnMouseDoubleClick(e);
    }

    private void refreshDockedControls()
    {
      foreach (Control c in this.Controls)
      {
        if (c.Dock != DockStyle.None)
        {
          DockStyle oldDockstyle = c.Dock;
          c.Dock = DockStyle.None;
          c.Dock = oldDockstyle;
        }
      }

      Invalidate();
    }

    protected override void OnResize(EventArgs eventargs)
    {
      base.OnResize(eventargs);

      ComputeButtonOffsets();
      Invalidate();
    }

    #endregion

    public int GetTabAtPosition(int xpos, int ypos)
    {
      if (DisplayMode == CollapsingPanelDisplayMode.HeaderAndTab ||
          DisplayMode == CollapsingPanelDisplayMode.TabsOnly ||
          DisplayMode == CollapsingPanelDisplayMode.HeaderWithTabs ||
          (xpos > headerHeight + Padding.Top && ypos < headerHeight + Padding.Top + tabsHeight))
      {
        int x = (DisplayMode == CollapsingPanelDisplayMode.HeaderWithTabs) ?
          (disableExpansion ? 0 : 32) : 8;

        for (int i = 0; i < tabCount; i++)
        {
          if (i >= cachedTabWidths.Count)
            break;

          if (cachedTabWidths[i] == 0)
            continue;

          if (xpos >= x && xpos < x + cachedTabWidths[i])
          {
            return i;
          }

          x += cachedTabWidths[i];
        }
      }
      return -1;
    }

    /// <summary>
    /// Adds a new custom button to the header with the given image.
    /// </summary>
    /// <returns>
    /// The index of the button in the list of custom buttons. Can be used in the click handler.
    /// </returns>
    public int AddCustomButton(int width, int height, Bitmap bitmap, bool active)
    {
      CustomButton button = new CustomButton(bitmap, new Rectangle(0, 0, width, height));
      customButtons.Add(button);
      if (active)
        activeCustomButton = button;
      ComputeButtonOffsets();

      return customButtons.Count - 1;
    }

    /// <summary>
    /// This version allows to specify a resource identifier which is then used to load the
    /// specific bitmap.
    /// </summary>
    public int AddCustomButton(int width, int height, String id, bool active)
    {
      object obj = Resources.ResourceManager.GetObject(id, Resources.Culture);
      return AddCustomButton(width, height, (System.Drawing.Bitmap)(obj), active);
    }

    /// <summary>
    /// Removes all custom buttons.
    /// </summary>
    public void ClearCustomButtons()
    {
      activeCustomButton = null;
      customButtons.Clear();
      ComputeButtonOffsets();
    }

    /// <summary>
    /// Computes the left position of the most left button (either custom or action button) and
    /// stores it in a member for later use. Also the bounding box for all custom buttons is updated
    /// to reflect the real position.
    /// </summary>
    private void ComputeButtonOffsets()
    {
      // For now we don't need any vertical adjustment, just the horizontal offset for all buttons.
      buttonOffset = 0;
      if (displayTabActionButtons)
        buttonOffset = Resources.collapsing_panel_header_tab_add.Width +
          Resources.collapsing_panel_header_tab_del.Width;
      foreach (CustomButton button in customButtons)
        buttonOffset += button.boundingBox.Width;
      buttonOffset = Width - buttonOffset;

      // Do a second loop over all buttons and do the final update now that we have
      // the correct left offset.
      int offset = buttonOffset;
      foreach (CustomButton button in customButtons)
      {
        button.boundingBox.X = offset;
        offset += button.boundingBox.Width;
      }
    }

    public void AdjustExpandedHeight(int requiredDisplayArea)
    {
      // getHeaderOffset() already includes the top padding value.
      if (!resizable)
      {
        computedExpandedHeight = requiredDisplayArea + GetHeaderOffset() + Padding.Bottom;
        if (expanded)
          Height = computedExpandedHeight;
      }
    }

    #region Properties

    public override LayoutEngine LayoutEngine
    {
      get
      {
        if (layoutEngine == null)
          layoutEngine = new ColumnLayout();
        return layoutEngine;
      }
    }

    public override Rectangle DisplayRectangle
    {
      get
      {
        // get the current header (+ tabs) height
        int h = GetHeaderOffset();

        // Modify current DisplayRectangle and subtract our header + tab area.
        Rectangle r = base.DisplayRectangle;
        r.Y += h;
        r.Height -= h;

        return r;
      }
    }

    /// <summary>
    /// Returns the needed space (height) for the header and tabs depending
    /// on the current display mode.
    /// </summary>
    /// <returns></returns>
    private int GetHeaderOffset()
    {
      int h;

      switch (DisplayMode)
      {
        case CollapsingPanelDisplayMode.HeaderAndTab:
          h = headerHeight + Padding.Top + headerSpace + tabsHeight;
          break;
        case CollapsingPanelDisplayMode.TabsOnly:
          h = headerHeightTabsOnly + headerSpace;
          break;
        default:
          if (style == CollapsingPanelStyle.Flat)
            h = headerHeight;
          else
            h = headerHeight + Padding.Top + headerSpace;
          break;
      }
      return h;
    }

    [Bindable(true), Category("Appearance"),
    Description("The mode the panel is displayed.")]
    public CollapsingPanelDisplayMode DisplayMode
    {
      get { return displayMode; }
      set
      {
        displayMode = value;
        Invalidate();
      }
    }

    [Bindable(true), Category("Appearance"),
    Description("The visual style of the panel.")]
    public CollapsingPanelStyle Style
    {
      get { return style; }
      set
      {
        if (style != value)
        {
          style = value;

          if (style == CollapsingPanelStyle.Flat &&
            displayMode == CollapsingPanelDisplayMode.HeaderWithTabs)
            headerHeight = 25;
          else
            headerHeight = 23;

          if (!Expanded)
            Height = headerHeight;
          else if (computedExpandedHeight > -1)
            Height = computedExpandedHeight;

          Invalidate();
        }
      }
    }

    [Bindable(true), Category("Appearance"),
    Description("The caption displayed in the panel header.")]
    public string HeaderCaption
    {
      get { return headerCaption; }
      set
      {
        headerCaption = value;
        Invalidate();
      }
    }

    [Bindable(true), Category("Appearance"),
    Description("The Font used for the panel header.")]
    public Font HeaderFont
    {
      get { return headerFont; }
      set
      {
        headerFont = value;
        Invalidate();
      }
    }

    [Bindable(true), Category("Appearance"),
    Description("The vertical space between the header and the client region.")]
    public int HeaderSpace
    {
      get { return headerSpace; }
      set
      {
        headerSpace = value;

        refreshDockedControls();
      }
    }

    [Bindable(true), Category("Appearance"),
    Description("The Font used for the tab header captions.")]
    public Font TabHeaderCaptionFont
    {
      get { return tabHeaderCaptionFont; }
      set
      {
        tabHeaderCaptionFont = value;
        Invalidate();
      }
    }

    [Bindable(true), Category("Appearance"),
    Description("The Font used for the tab header descriptions.")]
    public Font TabHeaderDescriptionFont
    {
      get { return tabHeaderDescriptionFont; }
      set
      {
        tabHeaderDescriptionFont = value;
        Invalidate();
      }
    }

    [Bindable(true), Category("Appearance"),
    Description("The ImageList holding the icons of the tab headers.")]
    public ImageList TabHeaderImageList
    {
      get { return tabHeaderImageList; }
      set { tabHeaderImageList = value; }
    }

    [Bindable(true), Category("Behavior"), DefaultValue(false),
    Description("Specifies the expanding state of the control.")]
    public bool Resizable
    {
      get { return resizable; }
      set
      {
        if (resizable != value)
        {
          resizable = value;
          Invalidate();
        }
      }
    }

    [Bindable(true), Category("Behavior"), DefaultValue(true),
    Description("Specifies the expanded state of the control.")]
    public bool Expanded
    {
      get { return expanded; }
      set
      {
        if (expanded != value)
        {
          expanded = value;
          if (!expanded)
          {
            if (resizable)
              manualHeight = Height;
            Height = headerHeight;
          }
          else
          {
            // Let the layout engine determine which height we need if not manually resizable.
            // It sets computedExpandedHeight to what is necessary.
            if (resizable)
              Height = manualHeight;
            else
            {
              PerformLayout();
              Height = computedExpandedHeight;
            }
          }
          Invalidate();
        }
      }
    }

    [Bindable(true), Category("Behavior"), DefaultValue(false),
    Description("Removes the expension icon and deactivates the expension of doubleclick.")]
    public bool DisableExpansion
    {
      get { return disableExpansion; }
      set
      {
        if (value != disableExpansion)
        {
          disableExpansion = value;
          Invalidate();
        }
      }
    }

    [Bindable(true), Category("Behavior"),
    Description("Specifies the index of the currently selected tab. The tab index is 1 based.")]
    public int SelectedTabIndex
    {
      get { return selectedTabIndex; }
      set
      {
        if (TabChanging != null)
          if (!TabChanging(this, value))
            return;

        if (selectedTabIndex != value)
        {
          selectedTabIndex = value;

          if (TabChanged != null)
            TabChanged(this, selectedTabIndex);

          Invalidate();
        }
      }
    }

    [Bindable(true), Category("Behavior"),
    Description("If set to true the tab action buttons will be displayed and activated.")]
    public bool DisplayTabActionButtons
    {
      get { return displayTabActionButtons; }
      set
      {
        if (value != displayTabActionButtons)
        {
          displayTabActionButtons = value;
          ComputeButtonOffsets();
          Invalidate();
        }
      }
    }

    [Bindable(true), Category("Behavior"),
    Description("If set to true the custom buttons will be displayed and activated.")]
    public bool DisplayCustomButtons
    {
      get { return displayCustomButtons; }
      set
      {
        if (value != displayCustomButtons)
        {
          displayCustomButtons = value;
          ComputeButtonOffsets();
          Invalidate();
        }
      }
    }

    [Bindable(true), Category("Behavior"),
    Description("Allows to activate a certain custom button by code.")]
    public int ActiveCustomButton
    {
      get
      {
        for (int i = 0; i < customButtons.Count; i++)
          if (customButtons[i] == activeCustomButton)
            return i;
        return -1;
      }
      set
      {
        if (activeCustomButton != null)
          Invalidate(activeCustomButton.boundingBox);

        // Be tolerant with indices out of bounds.
        if (value >= 0 && value < customButtons.Count)
        {
          activeCustomButton = customButtons[value];
          CustomButtonClicked(this, value);
          Invalidate(activeCustomButton.boundingBox);
        }
        else
          activeCustomButton = null;
      }
    }

    #endregion

      #region Control Events

      [Category("Action")]
      public event CustomButtonEvent CustomButtonClicked;

    [Category("Behavior")]
    public event TabInfoHandler TabInfoNeeded;

    [Category("Behavior")]
    public event TabCountHandler TabCountNeeded;

    [Category("Action")]
    public event TabChangingHandler TabChanging;

    [Category("Action")]
    public event TabChangedHandler TabChanged;

    [Category("Action")]
    public event EventHandler TabAddButtonClicked;

    [Category("Action")]
    public event EventHandler TabDelButtonClicked;

    [Category("Action")]
    public event MouseEventHandler TabDoubleClicked;

    [Category("Action")]
    public event MouseEventHandler TabHeaderMouseUp;

    #endregion
  }

  public class ColumnLayout : LayoutEngine
  {
    public override bool Layout(object container, LayoutEventArgs layoutEventArgs)
    {
      CollapsingPanel parent = container as CollapsingPanel;

      // Use DisplayRectangle so that parent.Padding is honored.
      Rectangle parentDisplayRectangle = parent.DisplayRectangle;
      Point nextControlLocation = parentDisplayRectangle.Location;

      // Loop over all controls and compute the overall size needed. Also move the controls to
      // their final place.
      int totalHeight = 0;
      foreach (Control c in parent.Controls)
      {
        // Only apply layout to visible controls.
        if (!c.Visible)
          continue;

        // Respect the margin of the control:
        // shift over the left and the top.
        nextControlLocation.Offset(c.Margin.Left, c.Margin.Top);

        // Set the location of the control and stretch it over the full width.
        if (c.Dock != DockStyle.Fill)
        {
          c.Location = nextControlLocation;
          c.Width = parentDisplayRectangle.Width - c.Margin.Left - c.Margin.Right;

          // Set the autosized controls to their autosized heights.
          if (c.AutoSize)
          {
            // Handle listview specially.
            if (c is ListView)
            {
              ListView listView = c as ListView;
              uint rawSize = Win32.SendMessage(listView.Handle, Win32.LVM_APPROXIMATEVIEWRECT, -1, 
                Win32.MakeLParam(ushort.MaxValue, ushort.MaxValue));

              ushort singleSize = Win32.HiWord(rawSize);

              // Nothing is perfect and neither is Listview's space computation. There is too much room in
              // large icon mode and a bit too few in tile mode. Correct that.
              switch (listView.View)
              {
                case View.LargeIcon:
                  singleSize -= 20;
                  break;
                case View.Tile:
                  singleSize += 5;
                  break;
              }
              if (listView.Groups.Count == 0)
                c.Height = (int) singleSize;
              else
                c.Height = listView.Groups.Count * (int) singleSize;
            }
            else
            {
              Size size = c.GetPreferredSize(new Size(parentDisplayRectangle.Width, int.MaxValue));
              c.Height = size.Height;
            }
          }
        }
        else
        {
          // If a control is set to fill the full area then we cannot layout it differently.
          c.SetBounds(parentDisplayRectangle.Left, parentDisplayRectangle.Top, parentDisplayRectangle.Width,
            parentDisplayRectangle.Height);
        }

        // Move X back to the display rectangle origin.
        nextControlLocation.X = parentDisplayRectangle.X;

        // Increment Y by the height of the control 
        // and the bottom margin.
        nextControlLocation.Y += c.Height + c.Margin.Bottom;

        totalHeight += c.Height + c.Margin.Top + c.Margin.Bottom;
      }

      parent.AdjustExpandedHeight(totalHeight);

      return false;
    }
  }

}
